# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

"""
@file __main__.py
@brief Main entry point for the BDNS API command line interface.
@details
This script provides a command line interface to interact with the BDNS API.
It allows users to fetch data from the API and save it to a file or print it to stdout.
@author: José María Cruz Lorite <josemariacruzlorite@gmail.com>
"""

import logging

from bdns.fetch.cli import app

if __name__ == "__main__":
    # Configure logging only when run as CLI, not when imported
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        force=True,  # Ensure configuration is applied even if basicConfig was called before
    )
    app()



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from datetime import date, datetime
from typing import List, Optional
from pathlib import Path

import click
import typer
import dateparser

from bdns.fetch.types import (
    Order,
    Ambito,
    Direccion,
    TipoAdministracion,
    DescripcionTipoBusqueda,
)


class DateParamType(click.ParamType):
    name = "date"

    def convert(self, value, param, ctx) -> date:
        dt = dateparser.parse(value)
        if dt is None:
            self.fail(f"Could not parse date: {value}", param, ctx)
        return dt.date()


DateType = DateParamType()


output_file: Optional[Path] = typer.Option(
    "-",
    "--output-file",
    "-o",
    help="File to save the output. '-' means standard output.",
    show_default=True,
)
max_retries: int = typer.Option(
    3,
    "--max-retries",
    "-mr",
    min=1,
    help="Maximum number of retries for failed requests.",
    show_default=True,
)
wait_time: int = typer.Option(
    2,
    "--wait-time",
    "-wt",
    min=1,
    help="Time to wait between retries in seconds.",
    show_default=True,
)
version: Optional[bool] = typer.Option(
    None,
    "--version",
    help="Show the version and exit.",
    is_flag=True,
)
num_pages: Optional[int] = typer.Option(
    1,
    "--num-pages",
    "-np",
    min=0,
    help="For paginated endpoints, the number of pages to fetch. If 0, fetches all pages.",
    show_default=True,
)
from_page: Optional[int] = typer.Option(
    0,
    "--from-page",
    "-fp",
    min=0,
    help="Page number to start fetching from. If 0, starts from the first page.",
    show_default=True,
)
pageSize: Optional[int] = typer.Option(
    10000,
    "--pageSize",
    "-ps",
    min=1,
    max=10000,
    help="Number of results per page. The maximum allowed is 10000.",
    show_default=True,
)
order: Optional[Order] = typer.Option(
    None,
    "--order",
    "-ord",
    help="Order of the results. Can be 'nivel1', 'nivel2', 'nivel3', 'codConcesion', 'numeroConvocatoria', 'convocatoria', 'descripcionCooficial', 'instrumento', 'urlBR', 'fechaConcesion', 'beneficiario', 'importe', 'ayudaEquivalente' or 'tieneProyecto'.",
    show_default=True,
)
direccion: Optional[Direccion] = typer.Option(
    None,
    "--direccion",
    "-d",
    help="Direction of the search. Can be 'asc' or 'desc'.",
    show_default=True,
)
vpd: Optional[str] = typer.Option(
    "GE",
    "--vpd",
    "-vpd",
    help="VPD portal ID.",
    show_default=True,
)
vpd_required: str = typer.Option(
    ...,
    "--vpd",
    "-vpd",
    help="VPD portal ID.",
)
descripcion: Optional[str] = typer.Option(
    None,
    "--descripcion",
    "-desc",
    help="Title or part of it, in Spanish or co-official language.",
    show_default=True,
)
descripcionTipoBusqueda: Optional[DescripcionTipoBusqueda] = typer.Option(
    None,
    "--descripcionTipoBusqueda",
    "-dtb",
    help="Type of search to perform on the title. 1 - all words, 2 - any of the words, 0 - exact phrase. Any other value should not be taken into account.",
    show_default=True,
)
numeroConvocatoria: Optional[str] = typer.Option(
    None,
    "--numeroConvocatoria",
    "-nconv",
    help="BDNS number of the call to search for.",
    show_default=True,
)
mrr: Optional[bool] = typer.Option(
    False,
    "--mrr",
    "-mrr",
    help="Indicates if the search is for the Recovery and Resilience Mechanism (MRR).",
    show_default=True,
)
fechaDesde: Optional[date] = typer.Option(
    None,
    "--fechaDesde",
    "-fd",
    click_type=DateType,
    metavar="DATE",
    help="Start date of the period indicated for the search. See https://github.com/scrapinghub/dateparser for supported formats.",
    show_default=True,
)
fechaHasta: Optional[date] = typer.Option(
    None,
    "--fechaHasta",
    "-fh",
    click_type=DateType,
    metavar="DATE",
    help="End date of the period indicated for the search. See https://github.com/scrapinghub/dateparser for supported formats.",
    show_default=True,
)
tipoAdministracion: Optional[TipoAdministracion] = typer.Option(
    None,
    "--tipoAdministracion",
    "-ta",
    help="Type of administrative body being searched for 'C' for State Administration, 'A' for Autonomous Community, 'L' for Local Entity and 'O' for other Bodies.",
    show_default=True,
)
organos: Optional[List[int]] = typer.Option(
    None,
    "--organos",
    "-org",
    help="List of identifiers of the administrative bodies.",
    show_default=True,
)
regiones: Optional[List[int]] = typer.Option(
    None,
    "--regiones",
    "-r",
    help="List of identifiers of the selected impact regions, separated by commas.",
    show_default=True,
)
tiposBeneficiario: Optional[List[int]] = typer.Option(
    None,
    "--tiposBeneficiario",
    "-tb",
    help="List of identifiers of the selected beneficiary types, separated by commas.",
    show_default=True,
)
tiposBeneficiario_str: Optional[List[str]] = typer.Option(
    None,
    "--tiposBeneficiario",
    "-tb",
    help="List of beneficiary type codes, separated by commas.",
    show_default=True,
)
instrumentos: Optional[List[int]] = typer.Option(
    None,
    "--instrumentos",
    "-ins",
    help="List of identifiers of the selected aid instruments, separated by commas.",
    show_default=True,
)
finalidad: Optional[int] = typer.Option(
    None,
    "--finalidad",
    "-f",
    help="Identifier of the purpose of the spending policy.",
    show_default=True,
)
ayudaEstado: Optional[str] = typer.Option(
    None,
    "--ayudaEstado",
    "-ae",
    help="SA Number - State aid reference (only for State aid).",
    show_default=True,
)
codConcesion: Optional[str] = typer.Option(
    None,
    "--codConcesion",
    "-cc",
    help="Code of the concession to search for.",
    show_default=True,
)
idDocumento: Optional[int] = typer.Option(
    None,
    "--idDocumento",
    "-iddoc",
    help="Identifier of the document to search for.",
    show_default=True,
)
idDocumento_required: int = typer.Option(
    ...,
    "--idDocumento",
    "-iddoc",
    help="Identifier of the document.",
)
nifCif: Optional[str] = typer.Option(
    None,
    "--nifCif",
    "-n",
    help="NIF/CIF of the beneficiary.",
    show_default=True,
)
beneficiario: Optional[str] = typer.Option(
    None,
    "--beneficiario",
    "-b",
    help="ID of the beneficiary.",
    show_default=True,
)
actividad: Optional[List[int]] = typer.Option(
    None,
    "--actividad",
    "-act",
    help="List of identifiers of the selected activities, separated by commas.",
    show_default=True,
)
id: Optional[int] = typer.Option(
    None,
    "--id",
    "-id",
    help="Identifier of the document to search for.",
    show_default=True,
)
id_required: int = typer.Option(
    ...,
    "--id",
    "-id",
    help="Identifier of the document.",
)
objetivos: Optional[List[int]] = typer.Option(
    None,
    "--objetivos",
    "-obj",
    help="List of identifiers of the objectives of the concession, separated by commas.",
    show_default=True,
)
producto: Optional[List[int]] = typer.Option(
    None,
    "--producto",
    "-p",
    help="List of identifiers of the selected products, separated by commas.",
    show_default=True,
)
codigoAdmin: Optional[str] = typer.Option(
    None,
    "--codigoAdmin",
    "-ca",
    help="Admin code of the body.",
    show_default=True,
)
codigoAdmin_required: str = typer.Option(
    ...,
    "--codigoAdmin",
    "-ca",
    help="Admin code of the organ.",
)
codigo: Optional[str] = typer.Option(
    None,
    "--codigo",
    "-c",
    help="Code of the administrative body.",
    show_default=True,
)
idAdmon: Optional[TipoAdministracion] = typer.Option(
    None,
    "--idAdmon",
    "-ida",
    help="Identifier of the administrative body.",
    show_default=True,
)
idAdmon_required: TipoAdministracion = typer.Option(
    ...,
    "--idAdmon",
    "-ida",
    help="Type of administrative body: C (State), A (Autonomous Community), L (Local Entity), O (Other)",
)
ambito: Optional[Ambito] = typer.Option(
    None,
    "--ambito",
    "-amb",
    help="Indicator of the area where the search will be conducted (Concessions (C), State Aid (A), de Minimis (M), Sanctions (S), Political Parties (P), Large Beneficiaries (G)).",
    show_default=True,
)
busqueda: Optional[str] = typer.Option(
    None,
    "--busqueda",
    "-bus",
    help="Filter for the description field, must have a minimum length of 3.",
    show_default=True,
)
idPersona: Optional[int] = typer.Option(
    None,
    "--idPersona",
    "-idp",
    help="Identifier of the person.",
    show_default=True,
)
reglamento: Optional[List[int]] = typer.Option(
    None,
    "--reglamento",
    "-reg",
    help="List of identifiers of the selected regulations, separated by commas.",
    show_default=True,
)
anios: Optional[List[int]] = typer.Option(
    None,
    "--anios",
    "-an",
    help="List of years in which they have been a large beneficiary.",
    show_default=True,
)
vigenciaDesde: Optional[date] = typer.Option(
    None,
    "--vigenciaDesde",
    "-vd",
    click_type=DateType,
    metavar="DATE",
    help="Start date of the validity of the strategic plan. See https://github.com/scrapinghub/dateparser for supported formats.",
    show_default=True,
)
vigenciaHasta: Optional[date] = typer.Option(
    None,
    "--vigenciaHasta",
    "-vh",
    click_type=DateType,
    metavar="DATE",
    help="End date of the validity of the strategic plan. See https://github.com/scrapinghub/dateparser for supported formats.",
    show_default=True,
)
numConv: Optional[str] = typer.Option(
    None,
    "--numConv",
    "-nc",
    help="Number of the call.",
    show_default=True,
)
numConv_required: str = typer.Option(
    ...,
    "--numConv",
    "-nc",
    help="Number of the call.",
)
idPES_required: int = typer.Option(
    ...,
    "--idPES",
    "-idpes",
    help="Identifier of the strategic plan.",
)
idPES: Optional[int] = typer.Option(
    None,
    "--idPES",
    "-idpes",
    help="Identifier of the strategic plan.",
    show_default=True,
)
codigo: str = typer.Option(
    ...,
    "--codigo",
    "-cod",
    help="Organ code.",
)

verbose_flag: bool = typer.Option(
    False,
    "--verbose",
    "-v",
    help="Enable verbose logging to show detailed HTTP requests and responses.",
    show_default=True,
)



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import json
import logging
from typing import Any, Dict, Generator, List
from datetime import date

import requests
from tqdm import tqdm
from tenacity import retry, stop_after_attempt, retry_if_exception_type, wait_fixed

from bdns.fetch.utils import (
    format_url,
    format_date_for_api_request,
    extract_option_values,
)
from bdns.fetch.endpoints import *
from bdns.fetch.types import (
    TipoAdministracion,
    Ambito,
    Order,
    Direccion,
    DescripcionTipoBusqueda,
)
from bdns.fetch import options

# Use a named logger for this module, don't configure at import time
logger = logging.getLogger(__name__)

# Add a NullHandler to prevent logging errors if no handlers are configured
logger.addHandler(logging.NullHandler())


class BDNSClient:
    """
    Client for interacting with the BDNS API programmatically.

    Provides configurable retry behavior for all BDNS API endpoints.
    """

    def __init__(self, max_retries: int = 3, wait_time: int = 2):
        """
        Initialize the BDNS client with configurable retry settings.

        Args:
            max_retries (int): Maximum number of retries for failed requests. Default: 3
            wait_time (int): Time to wait between retries in seconds. Default: 2
        """
        self.max_retries = max_retries
        self.wait_time = wait_time

    def _log_retry_attempt(self, retry_state):
        """Log retry attempts with instance-specific retry count."""
        exc = retry_state.outcome.exception()
        exc_type = type(exc).__name__ if exc else "None"
        exc_msg = str(exc) if exc else "No exception"

        logger.warning(
            f' Retrying due to {exc_type}: "{exc_msg}". '
            f"Attempt {retry_state.attempt_number} of {self.max_retries}."
        )

    def _create_retry_decorator(self):
        """Create a retry decorator with instance-specific settings."""
        return retry(
            stop=stop_after_attempt(self.max_retries),
            retry=retry_if_exception_type((requests.RequestException,)),
            wait=wait_fixed(self.wait_time),
            before_sleep=self._log_retry_attempt,
        )

    def _fetch_single_page(self, url: str) -> Dict[str, Any]:
        """
        Fetches data from a single page with error handling and retries.
        """
        retry_decorator = self._create_retry_decorator()

        @retry_decorator
        def fetch_with_retries():
            # Log the outgoing request
            logger.debug(f"HTTP REQUEST: GET {url}")

            import time

            start_time = time.time()

            response = requests.get(url, timeout=30)

            end_time = time.time()
            response_time = (end_time - start_time) * 1000  # Convert to milliseconds

            # Log response details
            logger.debug(
                f"HTTP RESPONSE: {response.status_code} {response.reason} - {response_time:.1f}ms"
            )
            logger.debug(f"Response Headers: {dict(response.headers)}")

            try:
                data = response.json()
            except ValueError:
                data = response.text

            # Log response content size and basic info
            content_size = len(response.text)
            logger.debug(f"Response Content-Length: {content_size} bytes")

            if isinstance(data, dict):
                if "content" in data and isinstance(data["content"], list):
                    logger.debug(f"Response contains {len(data['content'])} items")
                if "totalPages" in data:
                    logger.debug(f"Total pages available: {data['totalPages']}")
                if "number" in data:
                    logger.debug(f"Current page: {data['number']}")

            # Handle API errors
            if isinstance(data, dict) and "codigo" in data and "error" in data:
                logger.error(f"API Error Response: {data}")
                from bdns.fetch.exceptions import BDNSError

                tech_details = (
                    f"API error code {data['codigo']}: {data['error']} from {url}"
                )
                tech_details += f"\nResponse status: {response.status_code}"
                tech_details += f"\nResponse headers: {dict(response.headers)}"
                tech_details += f"\nFull response data: {data}"

                raise BDNSError(
                    message=f"API returned error: {data['error']}",
                    suggestion="Check your parameters and try again. Use --help for valid options.",
                    technical_details=tech_details,
                )

            if response.status_code != 200:
                logger.error(f"HTTP Error {response.status_code}: {response.reason}")
                logger.error(f"Response body: {data}")
                from bdns.fetch.exceptions import handle_api_error

                response_text = (
                    json.dumps(data) if isinstance(data, dict) else str(data)
                )
                raise handle_api_error(
                    response.status_code, url, response_text, dict(response.headers)
                )

            return data

        return fetch_with_retries()

    def _fetch_paginated(
        self,
        base_url: str,
        params: Dict[str, Any],
        from_page: int = 0,
        num_pages: int = 0,
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Synchronous generator for paginated data fetching.
        """
        # Fetch the first page to get total page count
        first_page_params = {**params, "page": from_page}
        first_page_url = format_url(base_url, first_page_params)

        try:
            first_response = self._fetch_single_page(first_page_url)
            total_pages = first_response.get("totalPages", 1)

            # Yield items from the first page
            content = first_response.get("content", [])
            if isinstance(content, list):
                for item in content:
                    yield item

            # Determine pages to fetch
            to_page = (
                total_pages
                if num_pages == 0
                else min(from_page + num_pages, total_pages)
            )

            # If there are more pages, fetch them sequentially
            if from_page + 1 < to_page:
                pages_to_fetch = list(range(from_page + 1, to_page))

                for page in tqdm(
                    pages_to_fetch,
                    desc=f"Fetching from page {from_page + 1} to page {to_page} out of {total_pages} pages",
                ):
                    page_params = {**params, "page": page}
                    page_url = format_url(base_url, page_params)
                    response = self._fetch_single_page(page_url)
                    content = response.get("content", [])
                    if isinstance(content, list):
                        for item in content:
                            yield item

        except Exception as e:
            logger.error(f"Error in paginated fetch: {e}")
            raise

    def _fetch(
        self, url: str, params: Dict[str, Any] = None
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Fetches data from a single non-paginated endpoint with retries and error handling.
        """
        # Format URL with parameters
        if params:
            full_url = format_url(url, params)
        else:
            full_url = url

        # Use the synchronous single page fetch method
        data = self._fetch_single_page(full_url)

        # Yield individual items based on response structure
        if isinstance(data, list):
            # Direct list response
            for item in data:
                yield item
        elif isinstance(data, dict):
            if "content" in data and isinstance(data["content"], list):
                # Paginated response structure (but single page)
                for item in data["content"]:
                    yield item
            else:
                # Single object response
                yield data
        else:
            logger.warning(f"Unexpected response type: {type(data)}")
            yield data

    def _fetch_binary(self, url: str) -> bytes:
        """
        Synchronously fetches binary content from a URL using requests.
        """
        from bdns.fetch.exceptions import handle_api_response

        logger.debug(f"Starting binary fetch from: {url}")

        try:
            response = requests.get(url, timeout=30)

            logger.debug(
                f"Binary response: {response.status_code} - Content-Type: {response.headers.get('content-type', 'unknown')}"
            )

            if response.status_code == 200:
                content = response.content
                logger.debug(f"Binary content fetched: {len(content)} bytes")
                return content
            elif response.status_code == 204:
                logger.debug("No content returned (204)")
                return b""
            elif response.status_code == 404:
                logger.warning(f"Resource not found (404) for URL: {url}")
                return b""
            else:
                # Handle API errors using existing error handler
                raise handle_api_response(
                    response.status_code, url, response.text, dict(response.headers)
                )
        except requests.RequestException as e:
            logger.error(f"Request failed for binary fetch: {e}")
            raise

    @extract_option_values
    def fetch_actividades(
        self, vpd: str = options.vpd
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/actividades"""
        params = {"vpd": vpd}
        url = format_url(BDNS_API_ENDPOINT_ACTIVIDADES, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_sectores(self) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/sectores"""
        params = {}
        url = format_url(BDNS_API_ENDPOINT_SECTORES, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_regiones(
        self, vpd: str = options.vpd
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/regiones"""
        params = {"vpd": vpd}
        url = format_url(BDNS_API_ENDPOINT_REGIONES, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_finalidades(
        self, vpd: str = options.vpd
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/finalidades"""
        params = {"vpd": vpd}
        url = format_url(BDNS_API_ENDPOINT_FINALIDADES, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_beneficiarios(
        self, vpd: str = options.vpd
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/beneficiarios"""
        params = {"vpd": vpd}
        url = format_url(BDNS_API_ENDPOINT_TIPOS_BENEFICIARIOS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_instrumentos(
        self, vpd: str = options.vpd
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/instrumentos"""
        params = {"vpd": vpd}
        url = format_url(BDNS_API_ENDPOINT_INSTRUMENTOS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_reglamentos(
        self, vpd: str = options.vpd, ambito: Ambito = options.ambito
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/reglamentos"""
        params = {
            "vpd": vpd,
            "ambito": ambito.value if hasattr(ambito, "value") else ambito,
        }
        url = format_url(BDNS_API_ENDPOINT_REGLAMENTOS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_objetivos(
        self, vpd: str = options.vpd
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/objetivos"""
        params = {"vpd": vpd}
        url = format_url(BDNS_API_ENDPOINT_OBJETIVOS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_grandesbeneficiarios_anios(self) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/grandesbeneficiarios/anios"""
        params = {}
        url = format_url(BDNS_API_ENDPOINT_GRANDES_BENEFICIARIOS_ANIOS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_planesestrategicos(
        self, idPES: int = options.idPES_required
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/planesestrategicos"""
        params = {"idPES": idPES}
        url = format_url(BDNS_API_ENDPOINT_PLANESESTRATEGICOS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_organos(
        self,
        vpd: str = options.vpd,
        idAdmon: TipoAdministracion = options.idAdmon_required,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/organos"""
        params = {
            "vpd": vpd,
            "idAdmon": idAdmon.value if hasattr(idAdmon, "value") else idAdmon,
        }
        url = format_url(BDNS_API_ENDPOINT_ORGANOS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_organos_agrupacion(
        self,
        vpd: str = options.vpd,
        idAdmon: TipoAdministracion = options.idAdmon_required,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/organos/agrupacion"""
        params = {
            "vpd": vpd,
            "idAdmon": idAdmon.value if hasattr(idAdmon, "value") else idAdmon,
        }
        url = format_url(BDNS_API_ENDPOINT_ORGANOS_AGRUPACION, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_organos_codigo(
        self, codigo: str = options.codigo
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/organos/codigo"""
        params = {"codigo": codigo}
        url = format_url(BDNS_API_ENDPOINT_ORGANOS_CODIGO, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_organos_codigoadmin(
        self, codigoAdmin: str = options.codigoAdmin_required
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/organos/codigoAdmin"""
        params = {"codigoAdmin": codigoAdmin}
        url = format_url(BDNS_API_ENDPOINT_ORGANOS_CODIGO_ADMIN, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_convocatorias(
        self, vpd: str = options.vpd, numConv: str = options.numConv_required
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/convocatorias"""
        params = {"vpd": vpd, "numConv": numConv}
        url = format_url(BDNS_API_ENDPOINT_CONVOCATORIAS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_concesiones_busqueda(
        self,
        num_pages: int = options.num_pages,
        from_page: int = options.from_page,
        pageSize: int = options.pageSize,
        order: Order = options.order,
        direccion: Direccion = options.direccion,
        vpd: str = options.vpd,
        descripcion: str = options.descripcion,
        descripcionTipoBusqueda: DescripcionTipoBusqueda = options.descripcionTipoBusqueda,
        numeroConvocatoria: str = options.numeroConvocatoria,
        fechaDesde: date = options.fechaDesde,
        fechaHasta: date = options.fechaHasta,
        tipoAdministracion: TipoAdministracion = options.tipoAdministracion,
        organos: List[int] = options.organos,
        regiones: List[int] = options.regiones,
        nifCif: str = options.nifCif,
        beneficiario: int = options.beneficiario,
        instrumentos: List[int] = options.instrumentos,
        actividad: List[int] = options.actividad,
        finalidad: int = options.finalidad,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/concesiones/busqueda"""
        params = {
            "pageSize": pageSize,
            "order": order.value if order else None,
            "direccion": direccion.value if direccion else None,
            "vpd": vpd,
            "descripcion": descripcion,
            "tipoAdministracion": tipoAdministracion.value
            if tipoAdministracion
            else None,
            "descripcionTipoBusqueda": descripcionTipoBusqueda.value
            if descripcionTipoBusqueda
            else None,
            "fechaDesde": format_date_for_api_request(fechaDesde)
            if fechaDesde
            else None,
            "fechaHasta": format_date_for_api_request(fechaHasta)
            if fechaHasta
            else None,
            "organos": organos,
            "regiones": regiones,
            "nifCif": nifCif,
            "beneficiario": beneficiario,
            "instrumentos": instrumentos,
            "actividad": actividad,
            "finalidad": finalidad,
            "numeroConvocatoria": numeroConvocatoria,
        }
        # Remove None values to keep URL clean
        params = {k: v for k, v in params.items() if v is not None}

        yield from self._fetch_paginated(
            BDNS_API_ENDPOINT_CONCESIONES_BUSQUEDA,
            params=params,
            from_page=from_page,
            num_pages=num_pages,
        )

    @extract_option_values
    def fetch_ayudasestado_busqueda(
        self,
        num_pages: int = options.num_pages,
        from_page: int = options.from_page,
        pageSize: int = options.pageSize,
        order: Order = options.order,
        direccion: Direccion = options.direccion,
        vpd: str = options.vpd,
        descripcion: str = options.descripcion,
        descripcionTipoBusqueda: DescripcionTipoBusqueda = options.descripcionTipoBusqueda,
        numeroConvocatoria: str = options.numeroConvocatoria,
        codConcesion: str = options.codConcesion,
        fechaDesde: date = options.fechaDesde,
        fechaHasta: date = options.fechaHasta,
        tipoAdministracion: TipoAdministracion = options.tipoAdministracion,
        organos: List[int] = options.organos,
        regiones: List[int] = options.regiones,
        objetivos: List[int] = options.objetivos,
        nifCif: str = options.nifCif,
        beneficiario: int = options.beneficiario,
        instrumentos: List[int] = options.instrumentos,
        actividad: List[int] = options.actividad,
        ayudaEstado: str = options.ayudaEstado,
        reglamento: int = options.reglamento,
        finalidad: int = options.finalidad,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/ayudasestado/busqueda"""
        params = {
            "pageSize": pageSize,
            "order": order.value if order else None,
            "direccion": direccion.value if direccion else None,
            "vpd": vpd,
            "descripcion": descripcion,
            "tipoAdministracion": tipoAdministracion.value
            if tipoAdministracion
            else None,
            "descripcionTipoBusqueda": descripcionTipoBusqueda.value
            if descripcionTipoBusqueda
            else None,
            "fechaDesde": format_date_for_api_request(fechaDesde)
            if fechaDesde
            else None,
            "fechaHasta": format_date_for_api_request(fechaHasta)
            if fechaHasta
            else None,
            "organos": organos,
            "regiones": regiones,
            "objetivos": objetivos,
            "nifCif": nifCif,
            "beneficiario": beneficiario,
            "instrumentos": instrumentos,
            "actividad": actividad,
            "numeroConvocatoria": numeroConvocatoria,
            "codConcesion": codConcesion,
            "ayudaEstado": ayudaEstado,
            "reglamento": reglamento,
            "finalidad": finalidad,
        }
        params = {k: v for k, v in params.items() if v is not None}

        yield from self._fetch_paginated(
            BDNS_API_ENDPOINT_AYUDASESTADO_BUSQUEDA,
            params=params,
            from_page=from_page,
            num_pages=num_pages,
        )

    @extract_option_values
    def fetch_terceros(
        self,
        vpd: str = options.vpd,
        ambito: Ambito = options.ambito,
        busqueda: str = options.busqueda,
        idPersona: int = options.idPersona,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/terceros

        Args:
            vpd: Identificador del portal (e.g., "A02")
            ambito: Ámbito donde buscar - C, A, M, S, P, G
            busqueda: Filtro para descripción (mín. 3 caracteres)
            idPersona: Identificador de la persona
        """
        params = {
            "vpd": vpd,
            "ambito": ambito,
            "busqueda": busqueda,
            "idPersona": idPersona,
        }
        url = format_url(BDNS_API_ENDPOINT_TERCEROS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_convocatorias_busqueda(
        self,
        num_pages: int = options.num_pages,
        from_page: int = options.from_page,
        pageSize: int = options.pageSize,
        order: Order = options.order,
        direccion: Direccion = options.direccion,
        vpd: str = options.vpd,
        descripcion: str = options.descripcion,
        descripcionTipoBusqueda: DescripcionTipoBusqueda = options.descripcionTipoBusqueda,
        numeroConvocatoria: str = options.numeroConvocatoria,
        mrr: bool = options.mrr,
        fechaDesde: date = options.fechaDesde,
        fechaHasta: date = options.fechaHasta,
        tipoAdministracion: TipoAdministracion = options.tipoAdministracion,
        organos: List[int] = options.organos,
        regiones: List[int] = options.regiones,
        tiposBeneficiario: List[str] = options.tiposBeneficiario_str,
        instrumentos: List[int] = options.instrumentos,
        finalidad: int = options.finalidad,
        ayudaEstado: str = options.ayudaEstado,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/convocatorias/busqueda"""
        params = {
            "pageSize": pageSize,
            "order": order.value if order else None,
            "direccion": direccion.value if direccion else None,
            "vpd": vpd,
            "descripcion": descripcion,
            "descripcionTipoBusqueda": descripcionTipoBusqueda.value
            if descripcionTipoBusqueda
            else None,
            "numeroConvocatoria": numeroConvocatoria,
            "mrr": mrr,
            "fechaDesde": format_date_for_api_request(fechaDesde)
            if fechaDesde
            else None,
            "fechaHasta": format_date_for_api_request(fechaHasta)
            if fechaHasta
            else None,
            "tipoAdministracion": tipoAdministracion.value
            if tipoAdministracion
            else None,
            "organos": organos,
            "regiones": regiones,
            "tiposBeneficiario": tiposBeneficiario,
            "instrumentos": instrumentos,
            "finalidad": finalidad,
            "ayudaEstado": ayudaEstado,
        }
        params = {k: v for k, v in params.items() if v is not None}

        yield from self._fetch_paginated(
            BDNS_API_ENDPOINT_CONVOCATORIAS_BUSQUEDA,
            params=params,
            from_page=from_page,
            num_pages=num_pages,
        )

    @extract_option_values
    def fetch_convocatorias_ultimas(
        self, vpd: str = options.vpd
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/convocatorias/ultimas"""
        params = {"vpd": vpd}
        url = format_url(BDNS_API_ENDPOINT_CONVOCATORIAS_ULTIMAS, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_convocatorias_documentos(
        self, idDocumento: int = options.idDocumento_required
    ) -> bytes:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/convocatorias/documentos"""
        params = {"idDocumento": idDocumento}
        url = format_url(BDNS_API_ENDPOINT_CONVOCATORIAS_DOCUMENTOS, params)
        return self._fetch_binary(url)

    @extract_option_values
    def fetch_convocatorias_pdf(
        self, id: int = options.id_required, vpd: str = options.vpd_required
    ) -> bytes:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/convocatorias/pdf"""
        params = {"id": id, "vpd": vpd}
        url = format_url(BDNS_API_ENDPOINT_CONVOCATORIAS_PDF, params)
        return self._fetch_binary(url)

    @extract_option_values
    def fetch_grandesbeneficiarios_busqueda(
        self,
        num_pages: int = options.num_pages,
        from_page: int = options.from_page,
        pageSize: int = options.pageSize,
        order: Order = options.order,
        direccion: Direccion = options.direccion,
        vpd: str = options.vpd,
        anios: List[int] = options.anios,
        anio: str = None,  # Alias for backward compatibility
        nifCif: str = options.nifCif,
        beneficiario: int = options.beneficiario,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/grandesbeneficiarios/busqueda"""
        # Use anio if provided, otherwise use anios
        years_param = anio if anio is not None else anios

        params = {
            "vpd": vpd,
            "pageSize": pageSize,
            "order": order.value if order else None,
            "direccion": direccion.value if direccion else None,
            "anios": years_param,
            "nifCif": nifCif,
            "beneficiario": beneficiario,
        }
        params = {k: v for k, v in params.items() if v is not None}

        yield from self._fetch_paginated(
            BDNS_API_ENDPOINT_GRANDES_BENEFICIARIOS_BUSQUEDA,
            params=params,
            from_page=from_page,
            num_pages=num_pages,
        )

    @extract_option_values
    def fetch_minimis_busqueda(
        self,
        num_pages: int = options.num_pages,
        from_page: int = options.from_page,
        pageSize: int = options.pageSize,
        order: Order = options.order,
        direccion: Direccion = options.direccion,
        vpd: str = options.vpd,
        descripcion: str = options.descripcion,
        descripcionTipoBusqueda: DescripcionTipoBusqueda = options.descripcionTipoBusqueda,
        numeroConvocatoria: str = options.numeroConvocatoria,
        codConcesion: str = options.codConcesion,
        fechaDesde: date = options.fechaDesde,
        fechaHasta: date = options.fechaHasta,
        tipoAdministracion: TipoAdministracion = options.tipoAdministracion,
        organos: List[int] = options.organos,
        regiones: List[int] = options.regiones,
        nifCif: str = options.nifCif,
        beneficiario: int = options.beneficiario,
        instrumentos: List[int] = options.instrumentos,
        actividad: List[int] = options.actividad,
        reglamento: int = options.reglamento,
        producto: int = options.producto,
        finalidad: int = options.finalidad,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/minimis/busqueda"""
        params = {
            "pageSize": pageSize,
            "order": order.value if order else None,
            "direccion": direccion.value if direccion else None,
            "vpd": vpd,
            "descripcion": descripcion,
            "descripcionTipoBusqueda": descripcionTipoBusqueda.value
            if descripcionTipoBusqueda
            else None,
            "numeroConvocatoria": numeroConvocatoria,
            "codConcesion": codConcesion,
            "fechaDesde": format_date_for_api_request(fechaDesde)
            if fechaDesde
            else None,
            "fechaHasta": format_date_for_api_request(fechaHasta)
            if fechaHasta
            else None,
            "tipoAdministracion": tipoAdministracion.value
            if tipoAdministracion
            else None,
            "organos": organos,
            "regiones": regiones,
            "nifCif": nifCif,
            "beneficiario": beneficiario,
            "instrumentos": instrumentos,
            "actividad": actividad,
            "reglamento": reglamento,
            "producto": producto,
            "finalidad": finalidad,
        }
        params = {k: v for k, v in params.items() if v is not None}

        yield from self._fetch_paginated(
            BDNS_API_ENDPOINT_MINIMIS_BUSQUEDA,
            params=params,
            from_page=from_page,
            num_pages=num_pages,
        )

    @extract_option_values
    def fetch_planesestrategicos_busqueda(
        self,
        num_pages: int = options.num_pages,
        from_page: int = options.from_page,
        pageSize: int = options.pageSize,
        order: Order = options.order,
        direccion: Direccion = options.direccion,
        vpd: str = options.vpd,
        descripcion: str = options.descripcion,
        descripcionTipoBusqueda: DescripcionTipoBusqueda = options.descripcionTipoBusqueda,
        fechaDesde: date = options.fechaDesde,
        fechaHasta: date = options.fechaHasta,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/planesestrategicos/busqueda"""
        params = {
            "pageSize": pageSize,
            "order": order.value if order else None,
            "direccion": direccion.value if direccion else None,
            "vpd": vpd,
            "descripcion": descripcion,
            "descripcionTipoBusqueda": descripcionTipoBusqueda.value
            if descripcionTipoBusqueda
            else None,
            "fechaDesde": format_date_for_api_request(fechaDesde)
            if fechaDesde
            else None,
            "fechaHasta": format_date_for_api_request(fechaHasta)
            if fechaHasta
            else None,
        }
        params = {k: v for k, v in params.items() if v is not None}

        yield from self._fetch_paginated(
            BDNS_API_ENDPOINT_PLANESESTRATEGICOS_BUSQUEDA,
            params=params,
            from_page=from_page,
            num_pages=num_pages,
        )

    @extract_option_values
    def fetch_planesestrategicos_documentos(
        self, idDocumento: int = options.idDocumento_required
    ) -> bytes:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/planesestrategicos/documentos"""
        params = {"idDocumento": idDocumento}
        url = format_url(BDNS_API_ENDPOINT_PLANESESTRATEGICOS_DOCUMENTOS, params)
        return self._fetch_binary(url)

    @extract_option_values
    def fetch_planesestrategicos_vigencia(
        self, vpd: str = options.vpd, idPES: int = options.idPES_required
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/planesestrategicos/vigencia"""
        params = {"vpd": vpd, "idPES": idPES}
        url = format_url(BDNS_API_ENDPOINT_PLANESESTRATEGICOS_VIGENCIA, params)
        yield from self._fetch(url)

    @extract_option_values
    def fetch_partidospoliticos_busqueda(
        self,
        num_pages: int = options.num_pages,
        from_page: int = options.from_page,
        pageSize: int = options.pageSize,
        order: Order = options.order,
        direccion: Direccion = options.direccion,
        vpd: str = options.vpd,
        descripcion: str = options.descripcion,
        descripcionTipoBusqueda: DescripcionTipoBusqueda = options.descripcionTipoBusqueda,
        numeroConvocatoria: str = options.numeroConvocatoria,
        codConcesion: str = options.codConcesion,
        fechaDesde: date = options.fechaDesde,
        fechaHasta: date = options.fechaHasta,
        tipoAdministracion: TipoAdministracion = options.tipoAdministracion,
        organos: List[int] = options.organos,
        regiones: List[int] = options.regiones,
        nifCif: str = options.nifCif,
        beneficiario: int = options.beneficiario,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/partidospoliticos/busqueda"""
        params = {
            "pageSize": pageSize,
            "order": order.value if order else None,
            "direccion": direccion.value if direccion else None,
            "vpd": vpd,
            "descripcion": descripcion,
            "descripcionTipoBusqueda": descripcionTipoBusqueda.value
            if descripcionTipoBusqueda
            else None,
            "numeroConvocatoria": numeroConvocatoria,
            "codConcesion": codConcesion,
            "fechaDesde": format_date_for_api_request(fechaDesde)
            if fechaDesde
            else None,
            "fechaHasta": format_date_for_api_request(fechaHasta)
            if fechaHasta
            else None,
            "tipoAdministracion": tipoAdministracion.value
            if tipoAdministracion
            else None,
            "organos": organos,
            "regiones": regiones,
            "nifCif": nifCif,
            "beneficiario": beneficiario,
        }
        params = {k: v for k, v in params.items() if v is not None}

        yield from self._fetch_paginated(
            BDNS_API_ENDPOINT_PARTIDOSPOLITICOS_BUSQUEDA,
            params=params,
            from_page=from_page,
            num_pages=num_pages,
        )

    @extract_option_values
    def fetch_sanciones_busqueda(
        self,
        num_pages: int = options.num_pages,
        from_page: int = options.from_page,
        pageSize: int = options.pageSize,
        order: Order = options.order,
        direccion: Direccion = options.direccion,
        vpd: str = options.vpd,
        descripcion: str = options.descripcion,
        descripcionTipoBusqueda: DescripcionTipoBusqueda = options.descripcionTipoBusqueda,
        numeroConvocatoria: str = options.numeroConvocatoria,
        fechaDesde: date = options.fechaDesde,
        fechaHasta: date = options.fechaHasta,
        tipoAdministracion: TipoAdministracion = options.tipoAdministracion,
        organos: List[int] = options.organos,
        regiones: List[int] = options.regiones,
        nifCif: str = options.nifCif,
        beneficiario: int = options.beneficiario,
        instrumentos: List[int] = options.instrumentos,
        actividad: List[int] = options.actividad,
        finalidad: int = options.finalidad,
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetches data from https://www.infosubvenciones.es/bdnstrans/api/sanciones/busqueda"""
        params = {
            "pageSize": pageSize,
            "order": order.value if order else None,
            "direccion": direccion.value if direccion else None,
            "vpd": vpd,
            "descripcion": descripcion,
            "descripcionTipoBusqueda": descripcionTipoBusqueda.value
            if descripcionTipoBusqueda
            else None,
            "numeroConvocatoria": numeroConvocatoria,
            "fechaDesde": format_date_for_api_request(fechaDesde)
            if fechaDesde
            else None,
            "fechaHasta": format_date_for_api_request(fechaHasta)
            if fechaHasta
            else None,
            "tipoAdministracion": tipoAdministracion.value
            if tipoAdministracion
            else None,
            "organos": organos,
            "regiones": regiones,
            "nifCif": nifCif,
            "beneficiario": beneficiario,
            "instrumentos": instrumentos,
            "actividad": actividad,
            "finalidad": finalidad,
        }
        params = {k: v for k, v in params.items() if v is not None}

        yield from self._fetch_paginated(
            BDNS_API_ENDPOINT_SANCIONES_BUSQUEDA,
            params=params,
            from_page=from_page,
            num_pages=num_pages,
        )



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Created on Sat May 17 16:23:45 2025
Author: josemariacruzlorite@gmail.com
"""

BDNS_API_BASE_URL = "https://www.infosubvenciones.es/bdnstrans/api"
BDNS_API_ENDPOINT_CONVOCATORIAS_BUSQUEDA = f"{BDNS_API_BASE_URL}/convocatorias/busqueda"
BDNS_API_ENDPOINT_CONVOCATORIAS_ULTIMAS = f"{BDNS_API_BASE_URL}/convocatorias/ultimas"
BDNS_API_ENDPOINT_CONVOCATORIAS = f"{BDNS_API_BASE_URL}/convocatorias"
BDNS_API_ENDPOINT_CONVOCATORIAS_DOCUMENTOS = (
    f"{BDNS_API_BASE_URL}/convocatorias/documentos"
)
BDNS_API_ENDPOINT_CONVOCATORIAS_PDF = f"{BDNS_API_BASE_URL}/convocatorias/pdf"
BDNS_API_ENDPOINT_CONCESIONES_BUSQUEDA = f"{BDNS_API_BASE_URL}/concesiones/busqueda"
BDNS_API_ENDPOINT_AYUDASESTADO_BUSQUEDA = f"{BDNS_API_BASE_URL}/ayudasestado/busqueda"
BDNS_API_ENDPOINT_MINIMIS_BUSQUEDA = f"{BDNS_API_BASE_URL}/minimis/busqueda"
BDNS_API_ENDPOINT_GRANDES_BENEFICIARIOS_BUSQUEDA = (
    f"{BDNS_API_BASE_URL}/grandesbeneficiarios/busqueda"
)
BDNS_API_ENDPOINT_GRANDES_BENEFICIARIOS_ANIOS = (
    f"{BDNS_API_BASE_URL}/grandesbeneficiarios/anios"
)
BDNS_API_ENDPOINT_PARTIDOSPOLITICOS_BUSQUEDA = (
    f"{BDNS_API_BASE_URL}/partidospoliticos/busqueda"
)
BDNS_API_ENDPOINT_PLANESESTRATEGICOS_BUSQUEDA = (
    f"{BDNS_API_BASE_URL}/planesestrategicos/busqueda"
)
BDNS_API_ENDPOINT_PLANESESTRATEGICOS = f"{BDNS_API_BASE_URL}/planesestrategicos"
BDNS_API_ENDPOINT_PLANESESTRATEGICOS_DOCUMENTOS = (
    f"{BDNS_API_BASE_URL}/planesestrategicos/documentos"
)
BDNS_API_ENDPOINT_PLANESESTRATEGICOS_VIGENCIA = (
    f"{BDNS_API_BASE_URL}/planesestrategicos/vigencia"
)
BDNS_API_ENDPOINT_SANCIONES_BUSQUEDA = f"{BDNS_API_BASE_URL}/sanciones/busqueda"
BDNS_API_ENDPOINT_TERCEROS = f"{BDNS_API_BASE_URL}/terceros"
BDNS_API_ENDPOINT_REGIONES = f"{BDNS_API_BASE_URL}/regiones"
BDNS_API_ENDPOINT_FINALIDADES = f"{BDNS_API_BASE_URL}/finalidades"
BDNS_API_ENDPOINT_TIPOS_BENEFICIARIOS = f"{BDNS_API_BASE_URL}/beneficiarios"
BDNS_API_ENDPOINT_INSTRUMENTOS = f"{BDNS_API_BASE_URL}/instrumentos"
BDNS_API_ENDPOINT_REGLAMENTOS = f"{BDNS_API_BASE_URL}/reglamentos"
BDNS_API_ENDPOINT_SECTORES = f"{BDNS_API_BASE_URL}/sectores"
BDNS_API_ENDPOINT_ACTIVIDADES = f"{BDNS_API_BASE_URL}/actividades"
BDNS_API_ENDPOINT_OBJETIVOS = f"{BDNS_API_BASE_URL}/objetivos"
BDNS_API_ENDPOINT_ORGANOS = f"{BDNS_API_BASE_URL}/organos"
BDNS_API_ENDPOINT_ORGANOS_AGRUPACION = f"{BDNS_API_BASE_URL}/organos/agrupacion"
BDNS_API_ENDPOINT_ORGANOS_CODIGO = f"{BDNS_API_BASE_URL}/organos/codigo"
BDNS_API_ENDPOINT_ORGANOS_CODIGO_ADMIN = f"{BDNS_API_BASE_URL}/organos/codigoAdmin"
BDNS_API_ENDPOINT_MICROPORTAL_CONFIGURACION = (
    f"{BDNS_API_BASE_URL}/vpd/{{vpd}}/configuracion"
)
BDNS_API_ENDPOINT_ENLACES = f"{BDNS_API_BASE_URL}/enlaces"



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

"""
BDNS Fetch CLI: Command-line interface for BDNS data fetching.
"""

import typer
import functools
import logging
import click
from pathlib import Path

from bdns.fetch.utils import write_to_file
from bdns.fetch.client import BDNSClient
from bdns.fetch import options
from bdns.fetch import __version__


# Define a global BDNSClient instance with default parameters
bnds_client = None
app = typer.Typer()


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    output_file: Path = options.output_file,
    max_retries: int = options.max_retries,
    wait_time: int = options.wait_time,
    version: bool = options.version,
    verbose_flag: bool = options.verbose_flag,
):
    """
    BDNS Fetch - Base de Datos Nacional de Subvenciones (BDNS) CLI

    Fetch data from the Base de Datos Nacional de Subvenciones (BDNS).

    \b
    Examples:
      bdns-fetch --output-file organos.jsonl organos
      bdns-fetch --output-file convocatorias.jsonl convocatorias-busqueda --fechaDesde "2024-01-01"
      bdns-fetch --max-retries 5 --wait-time 1 ayudasestado-busqueda --descripcion "innovation"

    \b
    Official API: https://www.infosubvenciones.es/bdnstrans/api
    """
    # Handle version flag
    if version:
        typer.echo(f"bdns-fetch version {__version__}")
        raise typer.Exit()

    # If no subcommand is provided and version is not requested, show help
    if ctx.invoked_subcommand is None:
        typer.echo(ctx.get_help())
        raise typer.Exit()

    # Create configured client instance
    global bnds_client
    bnds_client = BDNSClient(
        max_retries=max_retries,
        wait_time=wait_time,
    )

    ctx.obj = {
        "output_file": output_file,
        "verbose": verbose_flag,
        "client": bnds_client,  # Store configured client in context
    }

    # Configure logging based on verbose flag
    if verbose_flag:
        # Set detailed logging only for bdns-fetch related loggers
        # Don't modify the root logger to avoid affecting other packages
        logging.getLogger("bdns.fetch").setLevel(logging.DEBUG)
        logging.getLogger("aiohttp.client").setLevel(logging.DEBUG)
        # Enable urllib3 logging for even more HTTP details
        logging.getLogger("urllib3.connectionpool").setLevel(logging.DEBUG)

        # Only set up a console handler if none exists for bdns.fetch
        bdns_logger = logging.getLogger("bdns.fetch")
        if not bdns_logger.handlers:
            handler = logging.StreamHandler()
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )
            handler.setFormatter(formatter)
            bdns_logger.addHandler(handler)
            bdns_logger.propagate = False  # Don't propagate to root logger
    else:
        # Keep default level for bdns.fetch logger only
        logging.getLogger("bdns.fetch").setLevel(logging.INFO)


def cli_wrapper(client_method_name):
    """
    Wrapper that executes a client method and writes the result to file.

    Args:
        client_method_name: The name of the client method to wrap

    Returns:
        A function that can be used as a Typer command
    """
    # Get the method signature from the global client instance
    client = BDNSClient()
    original_method = getattr(client, client_method_name)

    @functools.wraps(original_method)
    def wrapper(*args, **kwargs):
        ctx = click.get_current_context()
        output_file = ctx.obj["output_file"]

        # Call the method on the selected client
        client_method = getattr(bnds_client, client_method_name)
        data_generator = client_method(*args, **kwargs)
        write_to_file(data_generator, output_file)
        return None

    return wrapper


# Register all commands using method names
app.command("actividades")(cli_wrapper("fetch_actividades"))
app.command("sectores")(cli_wrapper("fetch_sectores"))
app.command("regiones")(cli_wrapper("fetch_regiones"))
app.command("finalidades")(cli_wrapper("fetch_finalidades"))
app.command("beneficiarios")(cli_wrapper("fetch_beneficiarios"))
app.command("instrumentos")(cli_wrapper("fetch_instrumentos"))
app.command("reglamentos")(cli_wrapper("fetch_reglamentos"))
app.command("objetivos")(cli_wrapper("fetch_objetivos"))
app.command("grandesbeneficiarios-anios")(
    cli_wrapper("fetch_grandesbeneficiarios_anios")
)
app.command("planesestrategicos")(cli_wrapper("fetch_planesestrategicos"))
app.command("organos")(cli_wrapper("fetch_organos"))
app.command("organos-agrupacion")(cli_wrapper("fetch_organos_agrupacion"))
app.command("organos-codigo")(cli_wrapper("fetch_organos_codigo"))
app.command("organos-codigoadmin")(cli_wrapper("fetch_organos_codigoadmin"))
app.command("convocatorias")(cli_wrapper("fetch_convocatorias"))
app.command("concesiones-busqueda")(cli_wrapper("fetch_concesiones_busqueda"))
app.command("ayudasestado-busqueda")(cli_wrapper("fetch_ayudasestado_busqueda"))
app.command("terceros")(cli_wrapper("fetch_terceros"))
app.command("convocatorias-busqueda")(cli_wrapper("fetch_convocatorias_busqueda"))
app.command("convocatorias-ultimas")(cli_wrapper("fetch_convocatorias_ultimas"))
app.command("convocatorias-documentos")(cli_wrapper("fetch_convocatorias_documentos"))
app.command("convocatorias-pdf")(cli_wrapper("fetch_convocatorias_pdf"))
app.command("grandesbeneficiarios-busqueda")(
    cli_wrapper("fetch_grandesbeneficiarios_busqueda")
)
app.command("minimis-busqueda")(cli_wrapper("fetch_minimis_busqueda"))
app.command("partidospoliticos-busqueda")(
    cli_wrapper("fetch_partidospoliticos_busqueda")
)
app.command("planesestrategicos-busqueda")(
    cli_wrapper("fetch_planesestrategicos_busqueda")
)
app.command("planesestrategicos-documentos")(
    cli_wrapper("fetch_planesestrategicos_documentos")
)
app.command("planesestrategicos-vigencia")(
    cli_wrapper("fetch_planesestrategicos_vigencia")
)
app.command("sanciones-busqueda")(cli_wrapper("fetch_sanciones_busqueda"))


if __name__ == "__main__":
    app()



# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Created on Sat May 17 16:23:45 2025
Author: josemariacruzlorite@gmail.com
"""

import sys
from contextlib import contextmanager
from datetime import datetime, date
from enum import Enum
import functools
import inspect
import json
from typing import Any, Dict, Generator
import requests

import typer
from typer.models import OptionInfo

from bdns.fetch.exceptions import handle_api_response


def format_date_for_api_request(value: date, output_format: str = "%d/%m/%Y"):
    """
    Formats a date for API requests.
    Args:
        date (datetime): The date to format.
        output_format (str): The format to use for the date. Default is "%d/%m/%Y".
    Returns:
        str: The formatted date as a string.
    """
    if value is None:
        return None
    if not isinstance(value, date):
        raise ValueError("The date must be a date object.")
    return value.strftime(output_format)


def format_url(url: str, query_params: dict):
    """
    Formats a URL with query parameters.
    Args:
        url (str): The base URL.
        query_params (dict): A dictionary containing the query parameters.
    Returns:
        str: The formatted URL with query parameters.
    """
    if not url.endswith("?"):
        url += "?"

    # Filter out None values and typer.OptionInfo objects, convert enums to values
    filtered_params = {}
    for key, value in query_params.items():
        if value is not None and not isinstance(value, typer.models.OptionInfo):
            # Convert enum values to their actual values
            if isinstance(value, Enum):
                filtered_params[key] = value.value
            else:
                filtered_params[key] = value

    url += "&".join([f"{key}={value}" for key, value in filtered_params.items()])
    return url


def api_request(url):
    """
    Fetches data from the BDNS API for concessions for a given date.
    Args:
        url (str): The URL to fetch data from.
    Returns:
        dict: A dictionary containing concessions data.
    Raises:
        BDNSAPIError: If the API request fails.
    """
    from bdns.fetch.exceptions import handle_api_error

    response = requests.get(url)

    if response.status_code == 200:
        result = response.json()
        if not result or (isinstance(result, list) and len(result) == 0):
            raise handle_api_response(200, url, response.text, dict(response.headers))
        return result
    else:
        raise handle_api_response(
            response.status_code, url, response.text, dict(response.headers)
        )


@contextmanager
def smart_open(file, *args, **kwargs):
    """
    Open a file, or use stdin/stdout if file is '-'.
    Passes all additional args/kwargs to open().
    """
    if str(file) == "-":
        sys.stdout.reconfigure(encoding="utf-8")
        yield sys.stdout
    else:
        with open(file, *args, **kwargs) as f:
            yield f


def extract_option_values(func):
    """
    Decorator that automatically extracts actual values from OptionInfo parameters.

    This allows methods to use options.* parameters in their signatures (for CLI help)
    while getting the actual default values when called programmatically.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Get the function signature
        sig = inspect.signature(func)

        # Bind arguments to get the full parameter mapping
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        # Extract values from all parameters
        for param_name, value in bound_args.arguments.items():
            if isinstance(value, OptionInfo):
                bound_args.arguments[param_name] = value.default

        return func(*bound_args.args, **bound_args.kwargs)

    return wrapper


def write_to_file(
    data_generator: Generator[Dict[str, Any], None, None], output_file: str = None
) -> None:
    """
    Streams data from a generator and writes each item to file as it comes.

    Args:
        data_generator: Generator that yields individual data items
        output_file: The output file path. If None, uses global _output_file or stdout
    """
    file_to_use = output_file or "-"

    with smart_open(file_to_use, "w", encoding="utf-8", buffering=1) as f:
        for item in data_generator:
            f.write(json.dumps(item, ensure_ascii=False) + "\n")
            f.flush()
